Package: gpboost
Type: Package
Title: Combining Tree-Boosting with Gaussian Process and Mixed Effects
        Models
Version: 1.5.4
Date: 2024-11-15
Authors@R: c(
	person("Fabio", "Sigrist", email = "fabiosigrist@gmail.com", role = c("aut", "cre")),
    person("Tim", "Gyger", role = c("aut")),
    person("Pascal", "Kuendig", role = c("aut")),
	person("Benoit", "Jacob", role = c("cph")),
	person("Gael", "Guennebaud", role = c("cph")),
	person("Nicolas", "Carre", role = c("cph")),
	person("Pierre", "Zoppitelli", role = c("cph")),
	person("Gauthier", "Brun", role = c("cph")),
	person("Jean", "Ceccato", role = c("cph")),
	person("Jitse", "Niesen", role = c("cph")),
	person("Other authors of Eigen for the included version of Eigen", role = c("ctb","cph")),
	person("Timothy A.", "Davis", role = c("cph")),
    person("Guolin", "Ke", role = c("ctb")),
    person("Damien", "Soukhavong", role = c("ctb")),
    person("James", "Lamb", role = c("ctb")),
	person("Other authors of LightGBM for the included version of LightGBM", role = c("ctb")),
    person("Microsoft Corporation", role = c("cph")),
    person("Dropbox, Inc.", role = c("cph")),
    person("Jay", "Loden", role = c("cph")),
    person("Dave", "Daeschler", role = c("cph")),
    person("Giampaolo", "Rodola", role = c("cph")),
    person("Alberto", "Ferreira", role = c("ctb")),
    person("Daniel", "Lemire", role = c("ctb")),
    person("Victor", "Zverovich", role = c("cph")),
    person("IBM Corporation", role = c("ctb")),
	person("Keith O'Hara", role = c("cph")),
	person("Stephen L. Moshier", role = c("cph")),
    person("Jorge Nocedal", role = c("cph")),
    person("Naoaki Okazaki", role = c("cph")),
    person("Yixuan Qiu", role = c("cph")),
    person("Dirk Toewe", role = c("cph"))
    )
Description: An R package that allows for combining tree-boosting with Gaussian process and mixed effects models. It also allows for independently doing tree-boosting as well as inference and prediction for Gaussian process and mixed effects models. See <https://github.com/fabsig/GPBoost> for more information on the software and Sigrist (2022, JMLR) <https://www.jmlr.org/papers/v23/20-322.html> and Sigrist (2023, TPAMI) <doi:10.1109/TPAMI.2022.3168152> for more information on the methodology.
Encoding: UTF-8
License: Apache License (== 2.0) | file LICENSE
URL: https://github.com/fabsig/GPBoost
BugReports: https://github.com/fabsig/GPBoost/issues
NeedsCompilation: yes
Biarch: true
Suggests: testthat
Depends: R (>= 3.5), R6 (>= 2.0)
Imports: data.table (>= 1.9.6), graphics, RJSONIO, Matrix (>= 1.1-0),
        methods, utils
SystemRequirements: C++17
RoxygenNote: 6.0.1
Packaged: 2024-11-15 17:05:39 UTC; whsigris
Author: Fabio Sigrist [aut, cre],
  Tim Gyger [aut],
  Pascal Kuendig [aut],
  Benoit Jacob [cph],
  Gael Guennebaud [cph],
  Nicolas Carre [cph],
  Pierre Zoppitelli [cph],
  Gauthier Brun [cph],
  Jean Ceccato [cph],
  Jitse Niesen [cph],
  Other authors of Eigen for the included version of Eigen [ctb, cph],
  Timothy A. Davis [cph],
  Guolin Ke [ctb],
  Damien Soukhavong [ctb],
  James Lamb [ctb],
  Other authors of LightGBM for the included version of LightGBM [ctb],
  Microsoft Corporation [cph],
  Dropbox, Inc. [cph],
  Jay Loden [cph],
  Dave Daeschler [cph],
  Giampaolo Rodola [cph],
  Alberto Ferreira [ctb],
  Daniel Lemire [ctb],
  Victor Zverovich [cph],
  IBM Corporation [ctb],
  Keith O'Hara [cph],
  Stephen L. Moshier [cph],
  Jorge Nocedal [cph],
  Naoaki Okazaki [cph],
  Yixuan Qiu [cph],
  Dirk Toewe [cph]
Maintainer: Fabio Sigrist <fabiosigrist@gmail.com>
Repository: CRAN
Date/Publication: 2024-11-15 17:40:02 UTC
Built: R 4.4.2; x86_64-apple-darwin15.6.0; 2024-11-17 04:07:12 UTC; unix
